#!/usr/local/bin/perl
#
# OMICHAIN
#
# XMM Pipeline Processing System OMICHAIN script
#
# Author: Jon Rainger (jfr@mssl.ucl.ac.uk)
#
# Description: This script runs the tasks required for to produce
#              PPS products from Imaging Mode ODFs.
#
# $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
#
#
#    MODIFICATIONS:
#    12/06/2000 - AMK - Removed call to OMPSF as none of the pipleline tasks use it's output.
#
#    12/06/2000 - AMK - Redirected sreen output to a file omichain_log as user will lose the output during a complete ODF
#
#    06/09/2000 - AMK - Ammended code so that the new style house keeping files are processed correctly.
#
#    06/09/2000 - AMK - Ammended code so that product names conform to specification.
#
#    29/09/2000 - AMk - Changed flat field product file name to tmp_flat_field as this file should NOT be a product.
#
#    11/10/2000 - AMK - Removed the die calls and replace with SAS::fatal calls.
#
#    03/11/2000 - CB  - Modified so that images with "BLOCKED" or "UNKNOWN" filters are not processed.
#                       Added call to ps2pdf.
#    21/01/2000 - CB - Modified so that "ommosaic" runs at the end to produce a composite image for each filter.
#    10/08/2001 - CB - Modified to process engineering 2 and 4 data and restructured.
#************************************************************************************************************************
use DAL;
use SAS;

sub omichain {
#/////////////////////////////////////////

    #************************************************
    # For program development message can be written
    # to a file "omichain_log"
    # REMEMBER TO BLANK OUT NEXT LINE BEFORE UPLOADING
    #***********************************************

#    &CreateLogFile;
    
    $taskname='OMICHAIN';
    $VERSION='1.27';
    $date=&date_time;
    $create = "Running SAS task $taskname V$VERSION $date";
    &Message("$create \n");


    #**************************************
    # Set up the directory paths for input
    # and output.
    #**************************************
    &SetUpDirectoryPaths;

    #***************************************************
    # Initialise the lists of engineering 2/4 data, etc.
    #***************************************************
    &InitialiseLists;

    #************************************************
    # Get a list of the files in the current directory
    #************************************************
    &GetListOfDataFiles;

    #************************************************** 
    # Check for the presence of the house keeping files
    #**************************************************
    &CheckForHouseKeepingFiles;

    #**********************************************
    # Run the pipeline- Loop over all observations.
    #**********************************************
    foreach $orb_key (keys(%orb_list))
    {

        #********************************************
        # Process the data for each orbit in the list
        #********************************************
        &ProcessOrbit;  

    }
  
    #********************************************
    # We don't want to keep any new image file
    # created from Engineering mode 2 data
    #********************************************
    &DeleteNewImageFiles;

    $date = &date_time;
       
    $message = "Finished running SAS task $taskname V$VERSION $date";
    &Message($message);

    

    #******************************
    # For test purposes
    #******************************

    if ($redirect eq 5)
    {
        close(STDOUT);
        close(STDERR);
    }

}

#***********************************************************
#
# Subroutine section
#
#***********************************************************
sub date_time {          # create the file creation datestring

    local $datestring;
    $datestring = `date -u`;
    chop $datestring;
    return $datestring;
}
#********************************************************************
# subroutine RunOmmosaic - 23.01.2000
# Produces a list of the SIMAGE2000 files for each filter present
# in the ODF data set, which are then fed into omosaic to produce 
# a composite image.
#********************************************************************
sub RunOmmosaic
{
    @SASFiles = <$inp_directory/*.SAS>;        

 
    $SASFile = shift @SASFiles;
 
   

    open (SASFILE, $SASFile) || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFILE"); 

 
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************

    $line = <SASFILE>;
       
    until (substr($line, 0, 2) eq "OM")
    {
        $line = <SASFILE>;
                                  
    } 

  
    while ($line = <SASFILE>)
    {
        
        if ($line =~ "/ Exposure Identifier")
        {                 
            $line = <SASFILE>;
            until($line =~ "FILTER =")
            {
	        $line = <SASFILE>;
           
            } 
            
            $pos = index($line, "//");

            $filterName = substr($line, 9, $pos - 10); # Get rid of the characters after the end of the filter
            if ($filterName eq "White")
            {
                $filterName = "WHITE";
	    }

           
            push(@ExposureFilterList, $filterName);
            #*********************************************
            # Add the filter to the list
            #*********************************************
            $numFilters = @FilterList;
            $exists = 0;
       
        
            for ($i = 0; $i <  $numFilters ; $i++)
            {
	   
                if ($filterName eq @FilterList[$i])
                {
               
                    $exists = 1;
                    break;

                }

            }

            if ($exists eq 0)
            {
                push(@FilterList, $filterName);
	    }
        }
    }
    close SASFILE;



    @Simage2000Files = <*SIMAGE2000*> ;    # Creates a list of FIT file names    
 

    $numFilters = @FilterList;
    $numExposures = @ExposureNoList;
  
   
    $numFiles = @Simage2000Files;
    

    #******************************************************
    # Compose a list of the SIMAGE2000 files for each filter
    # and then run ommosaic
    #  11.06.2001 - Need to exclude the high-resolution
    # central image from the list.
    #******************************************************
    
    for ($k = 0; $k < $numFilters; $k++)
    {    

        @osw_simage_list =();           # Initialise the file list    
        for ($j = 0; $j < $numFiles; $j++)
        {
               
            $text = @Simage2000Files[$j];       # The SIMAGE2000 file name
            $text1 = substr($text, 14, 3);      # The exposure number

            for ($i = 0; $i <  $numExposures; $i++)
            {
                if (@ExposureFilterList[$i] eq @FilterList[$k])
                { 
                    $text2 = substr(@ExposureNoList[$i], 0, 3);
            
                    if ($text1 eq $text2)  
                    {   
                        #**********************************************************
                        # Add the file to the list if it is not the high-resolution
                        # central image
                        #**********************************************************
                        if (&IsHighResCentralWindow($text) == 0)
                        {               
                            push(@osw_simage_list, $text);  # Add the file name to the list
                        }    
                    }            
                }
            }
	}
               
        #************************************************
        # ommosaic
        #************************************************          
         $temp= @osw_simage_list;
       	 
        if ($temp >  0) 
        {

            $text = "... ommosaic- mosaicing images for filter ". @FilterList[$k]."\n";
            &Message($text);
 
            $outputFileName = substr(@osw_simage_list[0], 0, 14) . &GetFilterCode(@FilterList[$k]) . "SIMAGE2000.FIT";
     
            $arg_list = "imagesets=\"" . join(" ", @osw_simage_list) . "\" mosaicedset=". $out_directory . '/'.$outputFileName;


            &Message("ommosaic $arg_list");
            system("ommosaic $arg_list") && SAS::fatal("Task Failure", "ommosaic");
        }
        
  
    }
}
#**************************************
# subroutine GetFilterCode
# Given a filter from the SAS file
# returns the filter code to be used
# in constructing the output file name.
#**************************************
sub GetFilterCode
{
    my($filter) = @_;  # The input filter name
    my($name);
    
    @OMFilters = ("BLOCKED", "V",  "MAGNIFIER", "U",  "B",  "WHITE", "UVW1", "UVM2", "UVW2");
    @OMFilterCodes = ("BL",  "VB", "MA",        "UB", "BB", "WL",    "W1",   "M2",   "W2");

    for ($i = 0; $i < 9; $i++)
    {
       
        if (@OMFilters[$i] eq $filter)
        {
            return @OMFilterCodes[$i];
        
	}
    }
   
    
}
#************************************
# Sub IsHighResCentralWindow
#************************************
sub IsHighResCentralWindow
{
    local ($fileName) = @_;
    local ($result);
    $result = 0;
             
    if($fileName =~ /OMS00/)
    {
       
        if($fileName =~ /IMAGE_1000/)
        {
            $result = 1;
       
        }     
    }    
    return $result;
}
#********************************************************************
# subroutine FindEngineering2ImageFiles - 02.08.2000
#********************************************************************
sub FindEngineering2ImageFiles
{
    $numExposures = @ExposureNoList;
  
    @ImageFiles = <$inp_directory/*IMI.FIT*>;        

    $numFiles = @ImageFiles;
   
    for ($i = 0; $i < $numExposures; $i++)
    {
        $exposureNo = substr(@ExposureNoList[$i], 0, 3);
                 
        @ExposureImageList =();           # Initialise the file list    
      
        for ($j = 0; $j < $numFiles; $j++)
        {
               
            $fileName = @ImageFiles[$j];       
            $pos = index($fileName, "IMI") - 24;

            $string = substr($fileName, $pos, 32);    # Get rid of the directory name

            $text = substr($string, 19, 3);           # The exposure number
    
            if ($text eq $exposureNo)
            {
                push(@ExposureImageList, $fileName);         
          
            }            
              
                  
        }
        $numIMIFiles = @ExposureImageList;
   
        $num = 0;
        $fileName = "";
        if ($numIMIFiles > 0)
        {
            for ($j = 0; $j < $numIMIFiles; $j++)
            {
        
                $fileName = @ExposureImageList[$j];       
                $pos = index($fileName, "IMI") - 24;
                $name = substr($fileName, $pos, 32);    # Get rid of the directory name               
                $pattern = "0" .  $j . "IMI";  
                $pos = index($name, $pattern);
                if ($pos > 0) 
                {
                    $num++;
                }
            }
            if ($num eq 4)
            {

                #*********************************************************
                # Combine these images into 1 file
                #********************************************************
                $fileName = @ExposureImageList[0];       
                $pos = index($fileName, "IMI") - 24;
                $name = substr($fileName, $pos, 32);    # Get rid of the directory name                          

                #*****************************************************************
                # The new image file is temporarily stored in the input directory 
                #*****************************************************************  

# NB inpdirectory needs changing to out_directory
                $outputFileName = $inp_directory . '/'. substr($name, 0, 23) . "0E2I.FIT"; 
  

                $arg_list = "imagesets=\"" . join(" ", @ExposureImageList) . "\"  outset="  . $outputFileName;

                &Message("omcomb $arg_list");
                
                system("omcomb $arg_list") && SAS::fatal("Task Failure", "omcomb");
                #***********************************************
                # Check to see if omengine2 ran successfully
                #***********************************************
                $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
                if ($exit_value eq 1)  
                { 
                    $type = 1;  # Assume data normal imaging data

		    $ExposureType{$exposureNo} = "NORMAL";
	                
     
                    &Message("The test in omcomb have failed - This exposure ($exposureNo) will be treated as normal imaging data");

               
                }      
                else
                {

                    #*******************************************************
                    # Remember that these IMI files should not be processed
                    #*******************************************************
                    for ($j = 0; $j < $numIMIFiles; $j++)
                    {
        

                        $fileName = @ExposureImageList[$j];       
                        $pos = index($fileName, "IMI") - 24;
                        $name = substr($fileName, $pos, 32);    # Get rid of the directory name               
         

                        push(@EngineeringModeIMIList, $name); 
                     }         
        	    		
                              
                        $ExposureType{$exposureNo} = "ENG2";
		
                        push(@E2IMIList, $outputFileName);   # A list of the combined image files           
        	
                }
            }
            else
            {
         
                $ExposureType{$exposureNo} = "NORMAL";
		       
            }

        }

    }
}
#********************************************************************
# subroutine FindEngineering4ImageFiles
#********************************************************************
sub FindEngineering4ImageFiles
{
    $numExposures = @ExposureNoList;
  
    @ImageFile = <$inp_directory/*E4I.FIT*>;        

    $numFiles = @ImageFile;
   
    if($numFiles eq 0) 
    {
        return;   
    }
    #************************************************************************
    # Create a list of engineering 4 image files, without the directory path
    #************************************************************************
    for ($i = 0; $i < $numFiles; $i++)
    {
        $fileName = @ImageFile[$i];       
        $pos = index($fileName, "E4I") - 24;
        $name = substr($fileName, $pos, 32);    # Get rid of the directory name                        
        push(@E4IMIList);                       # Add it to the list of engineering mode 4 data





        $fileName = $name;

        #*********************************************
        # Find out which exposure this file is
        #*********************************************
        $exposure = substr($fileName, 19, 3);      # The exposure number
    
        $found = 0;
        $j = 0;
        while ($j < $numExposures and $found eq 0)
        {
            
            $exposureNo = substr(@ExposureNoList[$j], 0, 3);
        
            if ($exposure eq  $exposureNo)
            { 
                $ExposureType{$exposureNo} = "ENG4";
		$found = 1;
                       
            }
            $j++;
        }

    }
}
#******************************************************
# Subroutine FillExposureList
# Obtains a list of the exposures in the summary file
#******************************************************  
sub FillExposureNoList
{
    @SASFiles = <$inp_directory/*.SAS>;        

 
    $SASFile = shift @SASFiles;
  
    open (SASFILE, $SASFile) || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFILE"); 
 
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************

    $line = <SASFILE>;
       
    until (substr($line, 0, 2) eq "OM")
    {
        $line = <SASFILE>;
                                  
    } 

    while ($line = <SASFILE>)
    {
        
        if ($line =~ "/ Exposure Identifier")
        {        
            $exposureNo = substr($line, 0, 4);
            push(@ExposureNoList, $exposureNo);         
             
        }
    }
    close SASFILE;
   

}
#*****************************************************
# FindFile - Searches the directory for the specified
# file
#
#*****************************************************
sub FindFile($)
{

    my ( $fileName) = @_;
    $length = length($fileName);
    $a = substr($fileName, %length- 7, 7);


    if (substr($fileName, %length- 7, 7) eq "E2I.FIT")
    {
        $process = "ENGINEERING";
        $number = 0;
        return $number;
    } 
     

    local $number = 0;

    for ($j = 0; $j < $numEModeFiles; $j++)
    {
             
        $file = $inp_directory . "/" . @EngineeringModeIMIList[$j];
        

        if ($fileName eq $file)
        {
            $number = 1;
            $process = "ENGINEERING";
       
            return $number;
        } 
    }
    $process = "NORMAL";

   
    return $number;    

}
#***************************************************************
# Subroutine DeleteNewImageFiles
# Renames the new image files created from each set of four
# engineering 2 mode image files.
#***************************************************************
sub DeleteNewImageFiles
{

    if ($out_directory ne $inp_directory)
    { 
        $num = @E2IMIList;   


        for ($j = 0; $j < $num; $j++)
        {
        
 
            $fileName = @E2IMIList[$j];       
            $pos = index($fileName, "E2I") - 24;
            $name = $out_directory . '/' . substr($fileName, $pos, 32);    # Get rid of the directory name               
    
      
            system("cp $fileName $name"); 

        }         
        	    		
        unlink (@E2IMIList);
    }
              


    #*********************************************
    # Move the flatfield file created by omflatgen
    # to the output directory and delete the one
    # in the input directory
    #*********************************************
    if ($flatFieldFileName ne "BLANK" and $out_directory ne $inp_directory) 
    {
        $name = $inp_directory . '/'. $flatFieldFileName;
        system("cp $flatFieldFileName  $name"); 
        unlink $flatFieldFileName;
    }




    #******************************************
    # Delete temp files
    #*****************************************
    chdir $out_directory;

    &DeleteFileList("tmp");

    #******************************************
    # Delete imi files
    #*****************************************
    &DeleteFileList("imi");
 
    #******************************************
    # Delete fit files
    #*****************************************
    &DeleteFileList("fit");
 

}
#***************************************
# DeleteFileList
#***************************************
sub DeleteFileList($)
{ 
    my ($name) = @_;
   
    @List = <*$name*> ;    # Creates a list of FIT file names    

    unlink (@List);
}

#**************************************************************
# Subroutine MissingImageFileWarning
#*************************************************************
sub MissingImageFileWarning($)
{ 

    my ($fileName) = @_;
  
    $value = $ExposureType{$exp_key};
    
    if ($value eq "ENG2" or value eq "ENG4" and $value ne "")
    {
        return;
    }

    &Message("- \n*** Warning ***");
    &Message("Window '$win_key' of exposure '$exp_key' in observation '$obs_key' is missing an image file.");
  

#    $im_file = "RRRR_" . $obs_key . "_OM" . $exp_type . $exp_key . $win_key . "IMI.FIT";
    $im_file = $orb_key . "_" . $obs_key . "_OM" . $exp_type . $exp_key . $win_key . "IMI.FIT";
	           

    &Message("****************************************");
    &Message("Please insert file $im_file into the odf.");
    &Message("****************************************");
}
#********************************************
# Subroutine CreateProductFileNames
#********************************************
sub CreateProductFileNames
{

    $osw_list_intermediary_wavelet_filename = "osw_list_wavelet_" . $obs_key . "_" . $exp_type . $exp_key . substr($win_key, 1,1) . ".fit";

    $number = substr($win_key, 1, 1) + 1;
    $flat_field_product_filename = $out_directory . "/tmp_flat_field";

    $modulo_8_product_filename = $out_directory . "/tmp_mod8_file.fit";

    #***********************************
    # Need to start at 01 rather than 00
    #***********************************
    
    if ($process eq "NORMAL")  # Rudi-5 case
    {
        $string1 = "SIMAGE";
        $string2 = "IMAGE_";
    }
    else                       # Engineering case
    {
        $string1 = "FSIMAG";
        $string2 = "FIMAG_";
    }
    

    $sky_coord_pps_product_filename = "P" . $obs_key . "OM" . $exp_type . $exp_key . $string1 . $number . "000.FIT";
 

    $image_pps_product_filename = "P" . $obs_key . "OM" . $exp_type . $exp_key . $string2 . $number . "000.FIT";
    $osw_list_intermediary_detect_filename = "P" . $obs_key . "OM" . $exp_type . $exp_key . "SWSRLI" . $number . "000.FIT";

}
#********************************************
# Subroutine CheckTHXFileExits
#********************************************
sub CheckTHXFileExists
{
    

    if (exists(${$orb_key . $obs_key . $exp_type . $exp_key . "00"}{THX})) 
    {
 

	$tracking_history_filename = $inp_directory . "/". ${${$orb_key . $obs_key . $exp_type . $exp_key . "00"}{"THX"}};
    }	 
    else    
    {
    
        #***********************************
        # Create a dummy name to fool OMPREP
        #***********************************
        $tracking_history_filename = "DUMMYTHX.FIT";

 
    }

}

#************************************************
# Subroutine ProcessImageFile
# Processes an OM imaging file
#***********************************************
sub ProcessImageFile($)
{ 

    my ($image_filename) = @_;   # The image file to be processed
  
    #******************************************
    # Set pipeline variables.
    #**************************
    &CreateProductFileNames;


    #************************************ 
    # Rename region file as not a product
    #************************************
    $region_file = $out_directory . "/tmp_region_file.asc";

    #***********************************
    # Produce image mode products.
    #***********************************

    &Message("... Running Image Mode Pipeline ...");
    #*********************************
    # omprep
    #*********************************
    &Message("... omprep...");


    $tmp_image_1 = $out_directory . "/tmp_image_1";
    $tmp_image_2 = $out_directory . "/tmp_image_2";
  
    $arg_list = "set=" . $image_filename .

                " nphset=" . $inp_directory . '/' . $non_periodic_hk_filename .
                " pehset=" . $inp_directory . '/' . $periodic_hk_filename .
                " wdxset=" . $inp_directory . '/' . $window_data_filename .
                " outset=" . $tmp_image_1;

    &Message("omprep $arg_list");

    #**************************************************************
    # Modified so that if omprep fails it just skips the observation,
    # but does not kill the chain (1.11.2000)
    #**************************************************************
    system("omprep $arg_list");

    $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
    if ($exit_value eq 1) # Failure in omprep 
    { 
         &Message("omprep has detected an error- this observation will not be processed");
         return;

    }     
    $rawImi = $out_directory . "/raw_imi.fit";
    system("cp $tmp_image_1 $rawImi");

    
    #******************************************
    # omcosflag
    #******************************************
    &Message("... omcosflag ...");

    $arg_list = "set=" . $tmp_image_1 .
                " thxset=" . $out_directory . "/tmp_tracking";

    &Message("omcosflag $arg_list");
    system("omcosflag $arg_list") && SAS::fatal("Task Failure","omcosflag failed.");
 
                
    $omcosflag = $out_directory . "/omcosflag.fit";
    system("cp $tmp_image_1 $omcosflag");

  
    #****************************************
    # omflatfield
    #****************************************
    &Message("... omflatfield ...");

    $arg_list = "set=" . $tmp_image_1 .
                " thxset=" . $out_directory . "/tmp_tracking" .
                " inorbitflatset="  . $in_orbit_flat_field_filename .
                " tsflatset=" . $flat_field_product_filename .
                " outset=" . $tmp_image_2;

    &Message("omflatfield $arg_list");

    system("omflatfield $arg_list") && SAS::fatal("Task Failure","omflatfield failed.");
               
    $omflatfield_imi = $out_directory . "/omflatfield_imi.fit";    
    $name =  $omflatfield_imi . substr($win_key, 1, 1);
    system("cp $tmp_image_2 $name");

    #***************************************
    # ommodmap
    #***************************************
    &Message("... ommodmap ...");

    $arg_list = "set=" . $tmp_image_2 .
                " mod8product=yes" .
		" mod8set=$modulo_8_product_filename" .
                " outset=" . $tmp_image_1 .
                " nsig=5" .
                " nbox=16";  # Was originally 3

    &Message("ommodmap $arg_list");
    system("ommodmap $arg_list") && SAS::fatal("Task Failure","ommodmap failed.");
                     
    $ommodmappedimi = $out_directory . "/ommodmappedimi.fit";
    system("cp $tmp_image_1 $ommodmappedimi");
    #**************************************************
    # omdetect
    #**************************************************
    &Message("... omdetect ...");
    $background="background.imi";
    $arg_list = "set=" . $tmp_image_1 .
                " mod8set=$modulo_8_product_filename" . 
                " flatset=$flat_field_product_filename" .
                " outputregionfile=True" .
                " background=" . $background .
                " regionfile=" . $region_file .
                " outset=" . $out_directory . '/' . $osw_list_intermediary_detect_filename;

 
    &Message("omdetect $arg_list");

    system("omdetect $arg_list") && SAS::fatal("Task Failure","omdetect failed.");

    #*********************************************** 
    # ommag
    #***********************************************
    &Message("... ommag ...");

    $arg_list = "set=" . $out_directory . '/' . $osw_list_intermediary_detect_filename .
                " wdxset=" . $inp_directory . '/' . $window_data_filename;

    &Message("ommag $arg_list");

    system("ommag $arg_list") && SAS::fatal("Task Failure","ommag failed.");

    #***********************************************
    # omatt
    #***********************************************
    &Message("... omatt ...");

    $arg_list = "set="  . $tmp_image_1 .
                " sourcelistset=" . $out_directory . '/' . $osw_list_intermediary_detect_filename .
                " ppsoswset=" . $out_directory . '/' . $sky_coord_pps_product_filename.
                " usecat=F";

    &Message("omatt $arg_list");

    system("omatt $arg_list") && SAS::fatal("Task Failure","omatt failed.");


    rename($tmp_image_1, $out_directory . '/' . $image_pps_product_filename);
    unlink($tmp_image_1);


    #**********************************************************
    # 12.6.2001
    # Add the file to the list if it is not the high-resolution
    # central image
    #**********************************************************
                  
    if (&IsHighResCentralWindow($osw_list_intermediary_detect_filename) == 0)
    {                   
     
        push(@osw_file_list, $out_directory . '/' . $osw_list_intermediary_detect_filename);
                  
    }

}
#***************************************
#
# Subroutine GetImageFileName
#***************************************
sub GetImageFileName
{ 
           
    #*****************************************
    # Check for enginering 2 mode image
    #*****************************************
    $temp = ${${$orb_key . $obs_key . $exp_type . $exp_key . $win_key}{"E2I"}};
    if ($temp ne "")
    { 

        $imageFileName = $inp_directory . '/' . $temp;
    
        $process = "ENGINEERING";
        
        return ($imageFileName);   # Engineering 2 mode image file
    }         


    #*****************************************
    # Check for engineering 4 mode image
    #*****************************************

    $temp = ${${$orb_key . $obs_key . $exp_type . $exp_key . $win_key}{"E4I"}};
    if ($temp ne "")
    {    
        $imageFileName = $inp_directory . '/' . $temp;
   
        return ($imageFileName);  # Engineering 4 mode image file
    }



    #*****************************************
    # Check for Rudi-5 image
    #*****************************************


    $temp = ${${$orb_key . $obs_key . $exp_type . $exp_key . $win_key}{"IMI"}};
    if ($temp ne "")
    {


        $imageFileName = $inp_directory . '/' . $temp;
    

        $process = "NORMAL";

        return ($imageFileName);
    }
 

    #*********************************************
    # Missing image file- issue a warning.
    #*********************************************
    &MissingImageFileWarning($imageFileName);
    return ("MISSING");
    
}
#************************************************************
# Subroutine GetListOfDataFiles
# Get a list of file names in the current working directory.
#************************************************************
sub GetListOfDataFiles
{
    opendir(INDIR, $inp_directory) || SAS::fatal("OMICHAIN","Unable to open $inp_directory"); 

    $periodic_file_flag = 0;
    $non_periodic_file_flag = 0;

    #******************************************************************************
    # Sort the file names according to observation, exposure, window and file type.
    #******************************************************************************
    while ( $file = readdir(INDIR))
    {
        if ($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})([0-9A-Z]{3}).FIT/i)
  
        {
            if ($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM[S|U|X][0-9a-zA-Z]{3}[0-9]{2}FFX.FIT/i) 
            {
                $ffx_list{"$1$2"} = "$inp_directory/$file";
            }
            else
            {
#***************************************************************************************************
# We need to filter out the housekeeping files for this loop. There is now only one PEH and NPH file 
# per ODF.
#
#
# We are now picking up the exposure type flag and this stored with the exposure number.
# This is necessary as we are required to insert the exposure type flag into the product file names. 
# This necessitates some further filtering below.
#****************************************************************************************************
 
	        if ($file !~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})PEH.FIT/i
	           && $file !~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})NPH.FIT/i)
	        {
                    #******************
                    # Orbit number list
                    #******************
	            $orb_list{$1} = "$1";
                    #****************** 
                    # Observation list.
                    #******************
	            ${${"$1"}{$2}} = "$1$2";
                    #**********************************************
                    # Exposure list.
                    # This contains a value like 'X123'. X is the exposure flag and 123 i the exposure number.
                    #**********************************************
	            ${${"$1$2"}{$3}} = "$1$2$3";
                    #*************
                    # Window list.
                    #************* 
	            ${${"$1$2$3"}{$4}} = "$1$2$3$4";
                    #************************************************
                    # File list.
                    # Check for E2I.FIT & E4I.FIT file
                    # and make sure it gets stored in the IMI list
                    #************************************************

                    if ($5 eq "E2I" or $5 eq "E4I")
                    {
                        $num = "IMI";
                        ${${"$1$2$3$4"}{$num}} = "$file";
           
                    }
                    else
                    {
                        ${${"$1$2$3$4"}{$5}} = "$file";
                    }

                 

                }
                else
                {
                #*****************************************
                # Pick-up the House Keeping files here.
                #*****************************************
 
	            if ($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})PEH.FIT/i)
	            {
		        $periodic_hk_filename = $file;

		        $periodic_file_flag = 1;
	            }

	            if($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})NPH.FIT/i)
	            {
		        $non_periodic_hk_filename = $file;

		        $non_periodic_file_flag = 1;
	            }
                }
            }
        }
    }
    closedir(INDIR);

}
#*********************************************
# GetFlatFieldImage
#*********************************************
sub GetFlatFieldImage
{

    if (exists($ffx_list{$orb_key . $obs_key}))
    {
        $in_orbit_flat_field_filename =  $ffx_list{$orb_key . $obs_key };
        $flatFieldFileName = "BLANK";
    
    }   
    else
    {
           
        $flatFieldFileName =  "P" . $obs_key . "OM" . $exp_type . "000" . "FLAFLD" . "0000.FIT";

        $arg_list = "outset=". $inp_directory . '/' . $flatFieldFileName;

        $message = "omflatgen $arg_list";
        SAS::message($SAS::AppMsg, $SAS::SparseMsg, $message);

        #********************************
        # Create an artificial one
        #********************************

        system("omflatgen $arg_list") && SAS::fatal("Task Failure", "omflatgen");

        $ffx_list{"$1$2"} = $flatFieldFileName;
        $in_orbit_flat_field_filename = $inp_directory . '/' . $flatFieldFileName;  
      
    }
  
    return ($in_orbit_flat_field_filename);

}
#****************************************************
# Subroutine ProcessExposure
#****************************************************
sub ProcessExposure
{

    #********************************************************************************************** 
    # We must take out the first character of the exposure key as it now contains the exposure flag
    #**********************************************************************************************
    $exp_type = substr($exposure_key, 0, 1);
    $exp_key = substr($exposure_key, 1, 3);
	
    &Message("Checking ORBIT $orb_key OBSERVATION $obs_key EXPOSURE $exp_key: \n");
    &Message("[THX][WDX][PROCESSING]");
    #**************************   
    # Check consistency of odf.
    #**************************
 
    &CheckTHXFileExists;

   
   
    #****************************
    # Check consistancy of odf.
    #
    # We need only check for the presence of the WDX and THX files here.
    #
    # We now have a true or dummy tracking history file name
    #****************************

    $window_data_filename = ${${$orb_key . $obs_key . $exp_type . $exp_key . "00"}{"WDX"}};

    if (-e  $window_data_filename eq 0)
    {
        &Message("-");
        &Message("*** Warning ***");
        &Message("Exposure '$exp_key' in observation '$obs_key' is missing a file.\n");
        $wdx_file = "RRRR_" . $obs_key . "_OM" . $exp_type .$exp_key . "00WDX.FIT";

        &Message("Please check file $wdx_file is in the odf.\n");
        return;
    }

    #**************************
    # Set pipeline variables.
    #**************************
  
    $tracking_history_plot_filename = $out_directory . "/P" . $obs_key . "OM" . $exp_type . $exp_key . "TSHPLT" . "0000.ps";

     
    $tracking_history_plot_filename1 = $out_directory . "/P" . $obs_key . "OM" . $exp_type . $exp_key . "TSHPLT" . "0000.PDF";

    $tracking_history_timeseries_filename = $out_directory . "/P" . $obs_key . "OM" . $exp_type . $exp_key . "TSTRTS0000.FIT";
    #************************************
    # Produce tracking history products.
    #************************************
    &Message("... Running Tracking pipeline ...");

    #********************************************
    #  omprep
    #********************************************
    &Message("... omprep ...");

    $tempTracking =  $out_directory . "/tmp_tracking";
    $arg_list = "set=" . $tracking_history_filename .
                " nphset=" . $inp_directory . '/' . $non_periodic_hk_filename .
                " pehset=" . $inp_directory . '/' . $periodic_hk_filename .
                " wdxset=" . $inp_directory . '/' . $window_data_filename .
                " outset=" . $tempTracking;                           

    system("omprep -v");
    &Message("... omprep  $arg_list");  

    system("omprep  $arg_list");

    $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
    if ($exit_value eq 1) # Failure in omprep 
    { 
        &Message("omprep has detected an error- this observation will not be processed");
        return;  
    }     
    $omprepThx =  $out_directory . "/omprep_thx.fit";
    system("cp $tempTracking $omprepThx");

    #************************************************************
    # omdrifthist
    #************************************************************
    &Message("... omdrifthist ...");

    $arg_list = "set=" . $out_directory . '/' . "tmp_tracking" .
                " plotfile=" .  $tracking_history_plot_filename;

    &Message("omdrifthist $arg_list");

    system("omdrifthist $arg_list") && SAS::fatal("Task Failue","omdrifthist failed.");
	     

    #*******************************************************************
    # Only run ps2df if tracking-history file exists
    #*******************************************************************
    if (exists(${$orb_key . $obs_key . $exp_type . $exp_key . "00"}{THX})) 
    {
        &Message("ps2pdf ...");
             
        system("ps2pdf  $tracking_history_plot_filename   $tracking_history_plot_filename1") && SAS::fatal("Task Failure","ps2pdf failed.");
     
    }

    #************************************************************
    # omthconv
    #************************************************************
    &Message("... omthconv ...");

    $arg_list = "thxset=" . $out_directory . '/' . "tmp_tracking" .
                " nphset=" .  $inp_directory . '/' . $non_periodic_hk_filename .
                " outset=" .  $tracking_history_timeseries_filename ;

    &Message("omthconv $arg_list");

    system("omthconv $arg_list") && SAS::fatal("Task Failure","omthconv failed.");
              

    #****************************************
    # Loop over all windows.
    #****************************************
    $win_list = ${${$exp_list}{$exposure_key}};


    foreach $win_key (keys(%{$win_list}))
    {
                  
                     
        #******************************************************************
        # Check to see if this image file is in the list of engineering mode
        # 2/4 image files- don't process if it is.
        #****************************************************************** 
        $imageFileName = &GetImageFileName;


        if ($imageFileName ne "MISSING" and &FindFile($imageFileName) eq 0)
        {
            &ProcessImageFile($imageFileName);


        }
  
    }

    unlink($out_directory . '/' . "tmp_tracking");

}
#***************************
# Subroutine ProcessOrbit
#***************************
sub ProcessOrbit
{
    $obs_list = $orb_list{$orb_key};

    foreach $obs_key (keys(%{$obs_list}))
    {
        $exp_list = ${${$obs_list}{$obs_key}};

        &Message("           Observation $obs_key ");
        
        #*************************************************
        # Get the flat-field (create one if necessary)
        #*************************************************
        $in_orbit_flat_field_filename = &GetFlatFieldImage; 

        &Message("Using flatfield $in_orbit_flat_field_filename ");

        #*****************************************************
        # Loop over all exposures- if an unused filter is found
        # the exposure is not processed.
        #*****************************************************
        foreach $exposure_key (keys(%{$exp_list}))
        {

            #****************************************
            # Process the data for each exposure
            #****************************************

	    &ProcessExposure;


        }

        #**********************************
        # Set remaining pipeline variables.
        #**********************************
        $observation_source_list_pps_product_filename = "P" . $obs_key . "OM" . $exp_type . "000" . "OBSMLI" . "0000.FIT";

        #**********************************************
        # omsrclistcomb.
        #********************************************** 
        &Message("... omsrclistcomb ...");

        $arg_list = "sourcelistsets=\"" . join(" ", @osw_file_list) . "\"" .   
		     " nsigma=1" .    # nsigma was originally 1
                     " outset=" . $out_directory . '/' . $observation_source_list_pps_product_filename;

        &Message("omsrclistcomb $arg_list");

        system("omsrclistcomb $arg_list") && SAS::fatal("Task Failure","omsrclistcomb failed.");

        #************************************************
        # ommosaic
        #************************************************          
        &RunOmmosaic;
    }

}
#*******************************
# sub SetUpDirectoryPaths
#*******************************
sub SetUpDirectoryPaths
{

    $inp_directory = stringParameter("inpdirectory");
    if ($inp_directory eq "")
    {


        
        #*************************************************
        # If the SAS_ODF environment variable has been set
        # then set the input directory to it- otherwise
        # set the input directory to the current directory
        #*************************************************
        $input = $ENV{SAS_ODF};

     



        if ($input eq "")
        {
            $inp_directory = &GetCurrentDirectory;
            $ENV{SAS_ODF} = $inp_directory;
    
        }
        else
        {


            #*************************************************
            # If SAS_ODF is set to a .SAS file then need
            # to get the ODF directory from it
            #************************************************

            $pos = index($input, "SAS");
            if ($pos > 0) 
            {

		$inp_directory = &GetODFDirectory($input);
            }
            else
            {


                $inp_directory = $input;
            }
        }

    }
   

    $out_directory = stringParameter("outdirectory");


    if ($out_directory eq "")
    {
	$out_directory = &GetCurrentDirectory;
    }
      
    &Message("inpdirectory = $inp_directory");
    &Message("outdirectory = $out_directory");

 
}
#**********************************
# sub CheckForHouseKeepingFiles
#**********************************
sub CheckForHouseKeepingFiles
{
    if($periodic_file_flag == 0)
    {
        &Message("The ODF does NOT contain a Periodic Housekeeping file. Please insert into the ODF. ");
        SAS::fatal("Task Failure","Incomplete ODF.");
    }

    if($non_periodic_file_flag == 0)
    {
        &Message("The ODF does NOT contain a Non Periodic Housekeeping file. Please insert into the ODF.");
        SAS::fatal("Task Failure","incomplete ODF.");
    }
}

#**************************************
# Subroutine InitialiseLists
#**************************************
sub InitialiseLists
{
    @FilterList;                   # Stores a list of the filters in the summary file
    @ExposureNoList;               # Stores a list of the exposure numbers
    @ExposureFilterList;           # Stores a list of the filter for each exposure number
    @EngineeringModeIMIList;       # Stores the image files that are engineering mode 2 or 4 data
    @E2IMIList;                    # Stores the IMI files created by omengine2
    @E2IMIList =();
    @E4IMIList;                    # Stores the engineering 4 mode image files
    @E4IMIList =();

    %ExposureType = ();

    &FillExposureNoList;           # Stores a list of exposures in the summary file. 
    &FindEngineering2ImageFiles;   # Search for engineering 2 mode data.
    &FindEngineering4ImageFiles;   # Search for engineering 4 mode data.

    $numEModeFiles = @EngineeringModeIMIList;
    
    $numExposures = @ExposureNoList;
  
    foreach $key (keys (%ExposureType))
    {
        $value = $ExposureType{$key};
    }

    $numNewIMIFiles = @E2IMIList;   
}
#**************************************
# Subroutine Message
#**************************************
sub Message($)
{ 

    my ($message) = @_;  
    if ($redirect eq 5) 
    {
        print("$message \n");
    }
    else  
    {
        SAS::message($SAS::AppMsg, $SAS::SparseMsg, $message);
    }

}
#***************************************
# CreateLogFile
#**************************************
sub CreateLogFile
{
    $redirect = 5;
    open (STDOUT, ">omichain_log");
    open (STDERR, ">&STDOUT");
    select (STDERR);
}
#***********************************
# Subroutine GetCurrentDirectory
#**********************************
sub GetCurrentDirectory
{
    $command = "pwd";
    $curdir = `$command`;

    $leng = length($curdir);
    substr($curdir, $leng - 1) = "";    
    return $curdir;
}


#******************************
# Subroutine GetODFDirectory
# Retrieves the directory-path
# of the ODF data files.
#******************************
sub GetODFDirectory($)
{
    my ($SASFile) = @_;  
   
    open (SASFILE, $SASFile) || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFILE"); 
 
    $line = <SASFILE>;
        

    while ($line = <SASFILE>)
    {
        
        if ($line =~ "PATH")
        {        

            $length = length($line);
            $directory = substr($line, 5, $length - 6); 
            if ($directory eq "." or $directory eq "./")
            {
                $length = length($SASFile);
                $directory = substr($SASFile, 0, $length - 32);
            }
            close SASFILE;
            return $directory; 
             
        }
    }
    close SASFILE;
    SAS::fatal("OMICHAIN", "Unable to find the ODF directory path");

}










